from pathlib import Path

import neovim  # type: ignore

from tryp import List, Map

from trypnv import command, NvimStatePlugin, msg_command, json_msg_command

from proteome.plugins.core import (AddByParams, Show, Create, SetProject, Next,
                                   Prev, StageI, Save, RemoveByIdent, BufEnter,
                                   StageII, StageIII, StageIV, Clone)
from proteome.plugins.history import (HistoryPrev, HistoryNext, HistoryStatus,
                                      HistoryLog, HistoryBrowse,
                                      HistoryBrowseInput, HistorySwitch)
from proteome.main import Proteome
from proteome.nvim import NvimFacade
from proteome.logging import Logging


class ProteomeNvimPlugin(NvimStatePlugin, Logging):

    def __init__(self, vim: neovim.Nvim) -> None:
        super().__init__(NvimFacade(vim))
        self.pro = None  # type: Proteome
        self._initialized = False
        self._post_initialized = False

    def state(self):
        return self.pro

    @command()
    def proteome_reload(self):
        self.proteome_quit()
        self.proteome_start()
        self._post_startup()

    @command()
    def proteome_quit(self):
        if self.pro is not None:
            self.vim.clean()
            self.pro.stop()
            self.pro = None

    @command(sync=True)
    def proteome_start(self):
        config_path = self.vim.ppath('config_path')\
            .get_or_else(Path('/dev/null'))
        bases = self.vim.ppathl('base_dirs')\
            .get_or_else(List())\
            .map(Path)
        type_bases = self.vim.pd('type_base_dirs')\
            .get_or_else(Map())\
            .keymap(lambda a: Path(a).expanduser())\
            .valmap(List.wrap)
        plugins = self.vim.pl('plugins') | List()
        self.pro = Proteome(self.vim.proxy, Path(config_path), plugins, bases,
                            type_bases)
        self.pro.start()
        self.pro.send(StageI())

    @neovim.autocmd('VimEnter')
    def vim_enter(self):
        if not self._post_initialized:
            self.proteome_post_startup()

    @command()
    def proteome_post_startup(self):
        self._post_initialized = True
        if self.pro is not None:
            self.pro.send(StageII().at(1))
            self.pro.send(StageIII().at(1))
            self.pro.send(StageIV().at(1))
        else:
            self.log.error('proteome startup failed')

    @command()
    def pro_plug(self, plug_name, cmd_name, *args):
        self.pro.plug_command(plug_name, cmd_name, args)

    @msg_command(Create)
    def pro_create(self):
        pass

    @json_msg_command(AddByParams)
    def pro_add(self):
        pass

    @msg_command(RemoveByIdent)
    def pro_remove(self):
        pass

    @msg_command(Show)
    def pro_show(self):
        pass

    @msg_command(SetProject)
    def pro_to(self):
        pass

    @msg_command(Next)
    def pro_next(self):
        pass

    @msg_command(Prev)
    def pro_prev(self):
        pass

    @msg_command(Save)
    def pro_save(self):
        pass

    # TODO start terminal at root dir
    # @msg_command(Term)
    # def pro_term(self):
        # pass

    @neovim.autocmd('BufEnter')
    def buf_enter(self):
        if self._post_initialized:
            self.pro.send(BufEnter(self.vim.buffer.proxy))

    @json_msg_command(Clone)
    def pro_clone(self):
        pass

    @msg_command(HistoryPrev)
    def pro_history_prev(self):
        pass

    @msg_command(HistoryNext)
    def pro_history_next(self):
        pass

    @msg_command(HistoryStatus)
    def pro_history_status(self):
        pass

    @msg_command(HistoryLog)
    def pro_history_log(self):
        pass

    @msg_command(HistoryBrowse)
    def pro_history_browse(self):
        pass

    @msg_command(HistoryBrowseInput)
    def pro_history_browse_input(self):
        pass

    @msg_command(HistorySwitch)
    def pro_history_switch(self):
        pass

__all__ = ['ProteomeNvimPlugin']



from typing import Callable, Any
from contextlib import contextmanager
from pathlib import Path

from flexmock import flexmock  # type: ignore

from tek.test import temp_dir  # type: ignore

import tryp
import tryp.test
from tryp import may, Maybe, Just, List

from proteome.nvim import NvimFacade  # type: ignore
from proteome.project import Project  # type: ignore

from trypnv.nvim import Buffer


class MockNvimFacade(NvimFacade):

    def __init__(self):
        self.vars = {}
        super(MockNvimFacade, self).__init__(None)
        self.target = self

    @may
    def var(self, name: str) -> Maybe[str]:  # type: ignore
        v = self.vars.get(name)
        if v is None:
            self.log.error('variable not found: {}'.format(name))
        return v

    @property
    def buffer(self):
        return Buffer(self, self, self.prefix)

    @property
    def windows(self):
        return List()

    def switch_root(self, root):
        pass

    def async(self, f: Callable[['NvimFacade'], Any]):
        return f(self)

    @contextmanager
    def main_event_loop(self):
        yield None

    def cmd(self, *a, **kw):
        pass

    def reload_windows(self):
        pass


class Spec(tryp.test.Spec):

    def setup(self):
        super().setup()
        self.temp_projects = Path(temp_dir('projects'))
        self.history_base = Path(temp_dir('history'))

    def mk_project(self, name, tpe):
        root = temp_dir(str(self.temp_projects / tpe / name))
        return Project.of(name, Path(root), tpe=Just(tpe))

    def object_files(self, pro: Project):
        objdir = self.history_base / pro.fqn / 'objects'
        return list((objdir).iterdir())


class MockNvimSpec(Spec):

    def setup(self):
        super().setup()
        self.vim = MockNvimFacade()
        self.vim_mock = flexmock(self.vim)

__all__ = ('Spec', 'MockNvimSpec', 'MockNvimFacade')



from fn import _  # type: ignore

from trypnv.machine import may_handle, message
from trypnv.nvim import Buffer

from tryp import List
from tryp.lazy import lazy

from proteome.state import ProteomeComponent
from proteome.env import Env
from proteome.ctags import Ctags
from proteome.project import Project
from proteome.plugins.core import Save, Added, BufEnter, StageIV

Gen = message('Gen')
Kill = message('Kill')
CurrentBuffer = message('CurrentBuffer')


class Plugin(ProteomeComponent):

    @property
    def ctags(self) -> Ctags:
        return self._ctags  # type: ignore

    @lazy
    def _ctags(self):
        return Ctags(self.vim)

    @property
    def _tags_file_name(self):
        return '.tags'

    def _gen(self, pro: Project):
        self.ctags.gen(pro)

    @may_handle(StageIV)
    def stage_4(self, env: Env, msg):
        return Gen(), CurrentBuffer()

    @may_handle(Save)
    def save(self, env: Env, msg):
        return Gen()

    @may_handle(Gen)
    def gen(self, env: Env, msg):
        if self.ctags.ready:
            env.projects.projects\
                .filter(_.want_ctags)\
                .map(self._gen)
            self.ctags.exec()

    # TODO kill dangling procs
    @may_handle(Kill)
    def kill(self, env: Env, msg):
        pass

    @may_handle(BufEnter)
    def buf_enter(self, env, msg):
        self.set_buffer_tags(env, List(msg.buffer))

    @may_handle(Added)
    def added(self, env: Env, msg: Added):
        return CurrentBuffer()

    @may_handle(CurrentBuffer)
    def buffer(self, env: Env, msg):
        self.set_buffer_tags(env, List(self.vim.buffer))

    def set_buffer_tags(self, env: Env, bufs: List[Buffer]):
        files = env.all_projects.map(_.root / self._tags_file_name)
        bufs.foreach(lambda a: a.amend_optionl('tags', files))

__all__ = ['Gen', 'Plugin']
